import copy
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any
from typing import Self

import gql
import graphql
import httpx
import pydantic
from gql.transport import httpx as gql_httpx

DEFAULT_QUERY_TIMEOUT = 10

_ASGIApp = Callable[
    [
        MutableMapping[str, Any],
        Callable[[], Awaitable[MutableMapping[str, Any]]],
        Callable[[MutableMapping[str, Any]], Awaitable[None]],
    ],
    Awaitable[None],
]


class GQLQuery:
    def __init__(self):
        self.headers: dict[str, str] = {}
        self.upload_files: bool = False

    def __deepcopy__(self, memo: dict[int, Any] | None):
        q = self.__class__()
        q.headers = copy.deepcopy(self.headers, memo)
        q.upload_files = self.upload_files
        return q

    def with_headers(self, headers: dict[str, str]) -> Self:
        q = copy.deepcopy(self)
        q.headers = headers
        return q

    def with_file_uploads(self) -> Self:
        q = copy.deepcopy(self)
        q.upload_files = True
        return q


class GQLClient:
    def __init__(
        self,
        *,
        base_url: str,
        target_app: _ASGIApp | None = None,
        schema: str | Path,
        headers: dict[str, str] | None = None,
        query_timeout: int = DEFAULT_QUERY_TIMEOUT,
    ):
        self.base_url = base_url
        self.target_app = target_app
        self.headers = headers
        self.query_timeout = query_timeout

        if isinstance(schema, Path):
            with schema.open(encoding="utf-8") as f:
                schema = f.read()
        self.schema = graphql.build_ast_schema(graphql.parse(schema))

    async def query[T: pydantic.BaseModel](
        self,
        result_type: type[T],
        request: gql.GraphQLRequest,
        *,
        headers: dict[str, str] | None = None,
        upload_files: bool = False,
    ) -> T:
        """Execute a GraphQL query and return validated result.

        Args:
            result_type: Pydantic model class to validate the response against
            request: GraphQL request with document and optional variables
            headers: Optional HTTP headers to merge with client defaults
            upload_files: Whether to enable file upload support

        Returns:
            Validated instance of result_type with the query response data
        """
        httpx_transport = (
            httpx.ASGITransport(app=self.target_app) if self.target_app else None
        )
        gql_transport = gql_httpx.HTTPXAsyncTransport(
            url=self.base_url,
            transport=httpx_transport,
            headers=(self.headers or {}) | (headers or {}),
            timeout=self.query_timeout,
        )
        client = gql.Client(
            transport=gql_transport,
            schema=self.schema,
            execute_timeout=self.query_timeout,
            serialize_variables=True,
            parse_results=True,
        )
        async with client as session:
            result = await session.execute(request, upload_files=upload_files)
            return result_type.model_validate(result)


def serialize_var(value: Any) -> Any:
    match value:
        case list() | tuple():
            return [serialize_var(v) for v in value]
        case dict():
            return {k: serialize_var(v) for k, v in value.items()}
        case pydantic.BaseModel():
            return value.model_dump(mode="json", by_alias=True, exclude_unset=True)
        case _:
            return value



from pathlib import Path

import pytest
from pydantic import ValidationError
from pytest_httpserver import HTTPServer
from pytest_httpserver.httpserver import HandlerType

from iron_gql.conftest import generate_and_import
from iron_gql.conftest import prepare_workspace
from iron_gql.conftest import sample_app_context
from iron_gql.conftest import setup_test_server


async def test_generate_and_execute_queries(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type Query {
            user(id: ID!): User
        }

        type Mutation {
            updateUser(input: UpdateUserInput!): User
        }

        type User {
            id: ID!
            name: String!
        }

        input UpdateUserInput {
            id: ID!
            name: String!
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        get_user = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    id
                    name
                }
            }
            '''
        )

        update_user = api_gql(
            '''
            mutation UpdateUser($input: UpdateUserInput!) {
                updateUser(input: $input) {
                    id
                    name
                }
            }
            '''
        )
    """

    state = {"user-1": "Graph"}

    def resolve_user(_root, _info, *, id: str):
        name = state.get(id)
        if name is None:
            return None
        return {"id": id, "name": name}

    def resolve_update_user(_root, _info, **kwargs):
        input_data = kwargs["input"]
        user_id = str(input_data["id"])
        state[user_id] = input_data["name"]
        return {"id": user_id, "name": input_data["name"]}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {
            "Query": {"user": resolve_user},
            "Mutation": {"updateUser": resolve_update_user},
        },
    ) as (api, queries):
        read_query = queries.get_user.with_headers({"Authorization": "Bearer token"})
        initial = await read_query.execute(id="user-1")
        assert initial.user is not None
        assert initial.user.name == "Graph"

        mutation_input = api.UpdateUserInput(id="user-1", name="Morty")
        updated = await queries.update_user.execute(input=mutation_input)
        assert updated.update_user.name == "Morty"
        refreshed = await queries.get_user.execute(id="user-1")
        assert refreshed.user is not None
        assert refreshed.user.name == "Morty"


async def test_list_allows_null_elements(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type Query {
            numbers1: [Int]!
            numbers2: [Int!]
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        NUMBERS = api_gql(
            '''
            query Numbers {
                numbers1
                numbers2
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        api, queries = generate_and_import(tmp_path)

        httpserver.expect_request(
            "/graphql/",
            method="POST",
            handler_type=HandlerType.ONESHOT,
        ).respond_with_json({"data": {"numbers1": [1, None], "numbers2": [1, 2]}})

        api.API_CLIENT.base_url = httpserver.url_for("/graphql/")

        response = await queries.NUMBERS.execute()
        assert response.numbers_1 == [1, None]
        assert response.numbers_2 == [1, 2]

        httpserver.expect_request(
            "/graphql/",
            method="POST",
            handler_type=HandlerType.ONESHOT,
        ).respond_with_json({"data": {"numbers1": [1, 2], "numbers2": [1, None]}})

        with pytest.raises(ValidationError):
            await queries.NUMBERS.execute()


async def test_variable_defaults_optional(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type Query {
            posts(limit: Int = 5): [Int!]!
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_POSTS = api_gql(
            '''
            query GetPosts($limit: Int = 5) {
                posts(limit: $limit)
            }
            '''
        )
    """

    def resolve_posts(_root, _info, *, limit: int = 5):
        return list(range(limit))

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"posts": resolve_posts}},
    ) as (_, queries):
        default_result = await queries.GET_POSTS.execute()
        assert default_result.posts == [0, 1, 2, 3, 4]

        explicit_result = await queries.GET_POSTS.execute(limit=2)
        assert explicit_result.posts == [0, 1]



import importlib
from pathlib import Path

import graphql
import pytest
from pydantic import alias_generators

from iron_gql import runtime
from iron_gql.conftest import clear_sample_app_modules
from iron_gql.conftest import generate_api
from iron_gql.conftest import import_path
from iron_gql.conftest import prepare_workspace
from iron_gql.conftest import sample_app_context
from iron_gql.conftest import write_file
from iron_gql.generator import generate_gql_package


def test_generate_with_schema_outside_src(tmp_path: Path):
    workspace = tmp_path / "workspace"
    workspace.mkdir()
    schema_path = tmp_path / "schema.graphql"
    write_file(
        schema_path,
        """
        type Query {
            ping: String!
        }
        """,
    )

    write_file(workspace / "sample_app/__init__.py", "")
    write_file(workspace / "sample_app/gql/__init__.py", "")
    write_file(
        workspace / "sample_app/settings.py",
        "GRAPHQL_URL = 'http://testserver/graphql/'\n",
    )
    write_file(
        workspace / "sample_app/queries.py",
        """
        from sample_app.gql.api import api_gql

        ping = api_gql(
            '''
            query Ping {
                ping
            }
            '''
        )
        """,
    )

    with sample_app_context(workspace):
        changed = generate_gql_package(
            schema_path=schema_path,
            package_full_name="sample_app.gql.api",
            base_url_import="sample_app.settings:GRAPHQL_URL",
            scalars={"ID": "builtins:str"},
            to_camel_fn_full_name="pydantic.alias_generators:to_camel",
            to_snake_fn=alias_generators.to_snake,
            src_path=workspace,
        )
        assert changed is True

        module_path = workspace / "sample_app/gql/api.py"
        expected_schema_ref = schema_path.resolve().relative_to(
            workspace.resolve(), walk_up=True
        )
        generated = module_path.read_text(encoding="utf-8")
        assert f'Path("{expected_schema_ref}")' in generated

        clear_sample_app_modules()
        api_module = importlib.import_module("sample_app.gql.api")
        assert isinstance(api_module.API_CLIENT.schema, graphql.GraphQLSchema)


def test_duplicate_operations_raise(tmp_path: Path):
    schema = """
        type Query {
            user(id: ID!): User
        }

        type User {
            id: ID!
            name: String
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        first_query = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    id
                }
            }
            '''
        )

        second_query = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    id
                    name
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with (
        sample_app_context(tmp_path),
        pytest.raises(
            ValueError,
            match=r"^Cannot compile different GraphQL queries with same name",
        ),
    ):
        generate_api(tmp_path)


def test_nested_input_objects_missing(tmp_path: Path):
    schema = """
        type Query {
            ping: Boolean
        }

        type Mutation {
            updateUser(input: UpdateUserInput!): Boolean
        }

        input UpdateUserInput {
            id: ID!
            address: AddressInput
        }

        input AddressInput {
            street: String!
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        update_user = api_gql(
            '''
            mutation UpdateUser($input: UpdateUserInput!) {
                updateUser(input: $input)
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        changed = generate_api(tmp_path)
        assert changed is True

        clear_sample_app_modules()
        api_module = importlib.import_module("sample_app.gql.api")

        address = api_module.AddressInput(street="Main St")
        api_module.UpdateUserInput(id="u-1", address=address)


def test_invalid_query_reports_error(tmp_path: Path, caplog: pytest.LogCaptureFixture):
    schema = """
        type Query {
            user: String
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        BROKEN = api_gql(
            '''
            query Broken {
                missingField
            }
            '''
        )
        """,
        schema=schema,
    )

    with import_path(tmp_path):
        caplog.set_level("ERROR")
        changed = generate_api(tmp_path)
        assert changed is False
        assert "missingField" in caplog.text


def test_fragment_cycle_reports_error(tmp_path: Path, caplog: pytest.LogCaptureFixture):
    schema = """
        type Query {
            user: User
        }

        type User {
            id: ID!
            name: String!
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        fragment_a = api_gql(
            '''
            fragment A on User {
                id
                ...B
            }
            '''
        )

        fragment_b = api_gql(
            '''
            fragment B on User {
                name
                ...A
            }
            '''
        )

        get_user = api_gql(
            '''
            query GetUser {
                user {
                    ...A
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with import_path(tmp_path):
        caplog.set_level("ERROR")
        changed = generate_api(tmp_path)
        assert changed is False
        assert "Cannot spread fragment" in caplog.text
        assert not (tmp_path / "sample_app/gql/api.py").exists()


def test_input_type_dependency_ordering(tmp_path: Path):
    schema = """
        type Query {
            ping: Boolean
        }

        type Mutation {
            createOrder(input: OrderInput!): Boolean
        }

        input OrderInput {
            id: ID!
            item: ItemInput!
        }

        input ItemInput {
            sku: String!
            quantity: Int!
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        create_order = api_gql(
            '''
            mutation CreateOrder($input: OrderInput!) {
                createOrder(input: $input)
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        changed = generate_api(tmp_path)
        assert changed is True

        generated = (tmp_path / "sample_app/gql/api.py").read_text()
        item_pos = generated.find("class ItemInput")
        order_pos = generated.find("class OrderInput")
        assert item_pos != -1, "ItemInput class not found"
        assert order_pos != -1, "OrderInput class not found"
        assert item_pos < order_pos, "ItemInput must appear before OrderInput"

        clear_sample_app_modules()
        api_module = importlib.import_module("sample_app.gql.api")
        item = api_module.ItemInput(sku="ABC123", quantity=2)
        api_module.OrderInput(id="o-1", item=item)


def test_self_referential_input_type(tmp_path: Path):
    schema = """
        type Query {
            ping: Boolean
        }

        type Mutation {
            createTree(root: TreeNode!): Boolean
        }

        input TreeNode {
            value: String!
            children: [TreeNode]
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        create_tree = api_gql(
            '''
            mutation CreateTree($root: TreeNode!) {
                createTree(root: $root)
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        changed = generate_api(tmp_path)
        assert changed is True

        generated = (tmp_path / "sample_app/gql/api.py").read_text()
        assert 'children: "list[TreeNode | None] | None"' in generated

        clear_sample_app_modules()
        api_module = importlib.import_module("sample_app.gql.api")
        leaf = api_module.TreeNode(value="leaf")
        parent = api_module.TreeNode(value="parent", children=[leaf])
        assert parent.children[0].value == "leaf"


def test_input_enums_and_defaults(tmp_path: Path):
    schema = """
        type Query {
            ping: Boolean
        }

        type Mutation {
            update(input: UpdateInput!): Boolean
        }

        enum Status {
            ACTIVE
            INACTIVE
        }

        input ChildInput {
            code: String!
        }

        input UpdateInput {
            status: Status
            note: String
            child: ChildInput = { code: "X" }
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        update = api_gql(
            '''
            mutation Update($input: UpdateInput!) {
                update(input: $input)
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        changed = generate_api(tmp_path)
        assert changed is True

        generated = (tmp_path / "sample_app/gql/api.py").read_text()
        assert "type Status = Literal['ACTIVE', 'INACTIVE']" in generated
        assert "status: Status | None = None" in generated
        assert "note: str | None = None" in generated
        assert "child: \"ChildInput | None\" = {'code': 'X'}" in generated

        clear_sample_app_modules()
        api_module = importlib.import_module("sample_app.gql.api")
        update = api_module.UpdateInput(status="ACTIVE")
        assert update.child == {"code": "X"}
        serialized = runtime.serialize_var(update)
        assert serialized == {"status": "ACTIVE"}



from pathlib import Path

import pytest
from pytest_httpserver import HTTPServer

from iron_gql.conftest import generate_api
from iron_gql.conftest import import_path
from iron_gql.conftest import prepare_workspace
from iron_gql.conftest import sample_app_context
from iron_gql.conftest import setup_test_server


async def test_union_result_validation(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type Query {
            node(id: ID!): Node
            count: Int!
        }

        union Node = User | Admin

        type User {
            id: ID!
            name: String!
        }

        type Admin {
            id: ID!
            name: String!
            permissions: [String!]!
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        get_node_and_count = api_gql(
            '''
            query GetNodeAndCount($id: ID!) {
                node(id: $id) {
                    __typename
                    ... on User {
                        id
                        name
                    }
                    ... on Admin {
                        id
                        name
                        permissions
                    }
                }
                count
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        if id == "user-1":
            return {"__typename": "User", "id": id, "name": "Morty"}
        return {
            "__typename": "Admin",
            "id": id,
            "name": "Rick",
            "permissions": ["portal"],
        }

    def resolve_count(_root, _info):
        return 3

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node, "count": resolve_count}},
    ) as (_, queries):
        result = await queries.get_node_and_count.execute(id="user-1")
        assert result.node is not None
        assert result.count == 3


async def test_union_with_interface_fragment(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        interface Node {
            id: ID!
        }

        type User implements Node {
            id: ID!
            name: String!
        }

        type Admin implements Node {
            id: ID!
            permissions: [String!]!
        }

        union Actor = User | Admin

        type Query {
            actor(id: ID!): Actor
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_ACTOR = api_gql(
            '''
            query GetActor($id: ID!) {
                actor(id: $id) {
                    __typename
                    ... on Node {
                        id
                    }
                    ... on User {
                        name
                    }
                    ... on Admin {
                        permissions
                    }
                }
            }
            '''
        )
    """

    def resolve_actor(_root, _info, *, id: str):
        if id == "user-1":
            return {"__typename": "User", "id": id, "name": "Morty"}
        return {"__typename": "Admin", "id": id, "permissions": ["portal"]}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"actor": resolve_actor}},
    ) as (api, queries):
        user_result = await queries.GET_ACTOR.execute(id="user-1")
        assert isinstance(user_result.actor, api.GetActorResultActorUser)
        assert user_result.actor.id == "user-1"
        assert user_result.actor.name == "Morty"

        admin_result = await queries.GET_ACTOR.execute(id="admin-1")
        assert isinstance(admin_result.actor, api.GetActorResultActorAdmin)
        assert admin_result.actor.id == "admin-1"
        assert admin_result.actor.permissions == ["portal"]


async def test_interface_without_fragments(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        interface Node {
            id: ID!
        }

        type User implements Node {
            id: ID!
            name: String
        }

        type Post implements Node {
            id: ID!
            title: String
        }

        type Query {
            node(id: ID!): Node
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    id
                }
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        if id == "user-1":
            return {"__typename": "User", "id": id, "name": "Morty"}
        return {"__typename": "Post", "id": id, "title": "GraphQL 101"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node}},
    ) as (_, queries):
        result = await queries.GET_NODE.execute(id="user-1")
        assert result.node is not None
        assert result.node.id == "user-1"


async def test_interface_with_fragments(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        interface Node {
            id: ID!
        }

        type User implements Node {
            id: ID!
            name: String
        }

        type Post implements Node {
            id: ID!
            title: String
        }

        type Comment implements Node {
            id: ID!
            body: String
        }

        type Query {
            node(id: ID!): Node
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    __typename
                    id
                    ... on User {
                        name
                    }
                    ... on Post {
                        title
                    }
                }
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        if id == "user-1":
            return {"__typename": "User", "id": id, "name": "Morty"}
        if id == "post-1":
            return {"__typename": "Post", "id": id, "title": "GraphQL 101"}
        return {"__typename": "Comment", "id": id, "body": "First!"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node}},
    ) as (api, queries):
        user_result = await queries.GET_NODE.execute(id="user-1")
        assert isinstance(user_result.node, api.GetNodeResultNodeUser)
        assert user_result.node.name == "Morty"

        comment_result = await queries.GET_NODE.execute(id="comment-1")
        assert isinstance(comment_result.node, api.GetNodeResultNodeNode)
        assert comment_result.node.id == "comment-1"


async def test_nested_interface(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        interface Child {
            id: ID!
        }

        interface Node {
            id: ID!
            child: Child
        }

        type User implements Node {
            id: ID!
            child: Child
        }

        type Post implements Node {
            id: ID!
            child: Child
        }

        type Comment implements Child {
            id: ID!
        }

        type Query {
            node(id: ID!): Node
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    __typename
                    id
                    child {
                        id
                    }
                }
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        return {
            "__typename": "User",
            "id": id,
            "child": {"__typename": "Comment", "id": "child-1"},
        }

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node}},
    ) as (_, queries):
        result = await queries.GET_NODE.execute(id="user-1")
        assert result.node is not None
        assert result.node.child is not None
        assert result.node.child.id == "child-1"


async def test_interface_hierarchy(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        interface Node {
            id: ID!
        }

        interface Entity implements Node {
            id: ID!
            createdAt: String!
        }

        type User implements Entity & Node {
            id: ID!
            createdAt: String!
            name: String
        }

        type Post implements Node {
            id: ID!
            title: String
        }

        type Query {
            node(id: ID!): Node
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    __typename
                    id
                    ... on Entity {
                        createdAt
                    }
                }
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        if id == "user-1":
            return {
                "__typename": "User",
                "id": id,
                "createdAt": "2024-01-01",
                "name": "Morty",
            }
        return {"__typename": "Post", "id": id, "title": "GraphQL 101"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node}},
    ) as (api, queries):
        user_result = await queries.GET_NODE.execute(id="user-1")
        assert isinstance(user_result.node, api.GetNodeResultNodeUser)
        assert user_result.node.created_at == "2024-01-01"

        post_result = await queries.GET_NODE.execute(id="post-1")
        assert isinstance(post_result.node, api.GetNodeResultNodeNode)
        assert post_result.node.id == "post-1"


def test_interface_fragment_requires_typename(tmp_path: Path):
    schema = """
        interface Node {
            id: ID!
        }

        type User implements Node {
            id: ID!
            name: String
        }

        type Query {
            node(id: ID!): Node
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    id
                    ... on User {
                        name
                    }
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with (
        sample_app_context(tmp_path),
        pytest.raises(
            ValueError,
            match=r"Missing __typename in selection set for interface 'Node'",
        ),
    ):
        generate_api(tmp_path)


def test_invalid_interface_fragment_reports_error(
    tmp_path: Path, caplog: pytest.LogCaptureFixture
):
    schema = """
        interface Node {
            id: ID!
        }

        type User implements Node {
            id: ID!
            name: String
        }

        type Post {
            id: ID!
            title: String
        }

        type Query {
            node(id: ID!): Node
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        GET_NODE = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    __typename
                    id
                    ... on Post {
                        title
                    }
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with import_path(tmp_path):
        caplog.set_level("ERROR")
        changed = generate_api(tmp_path)
        assert changed is False
        assert "Post" in caplog.text
        assert "Node" in caplog.text


def test_union_fragment_requires_typename(tmp_path: Path):
    schema = """
        union SearchResult = User | Post

        type User {
            id: ID!
            name: String!
        }

        type Post {
            id: ID!
            title: String!
        }

        type Query {
            search(q: String!): SearchResult
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        SEARCH = api_gql(
            '''
            query Search($q: String!) {
                search(q: $q) {
                    ... on User {
                        id
                        name
                    }
                    ... on Post {
                        id
                        title
                    }
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with (
        sample_app_context(tmp_path),
        pytest.raises(
            ValueError,
            match=r"Missing __typename in selection set for union 'SearchResult'",
        ),
    ):
        generate_api(tmp_path)


def test_union_fragment_typename_in_variants(tmp_path: Path):
    schema = """
        union SearchResult = User | Post

        type User {
            id: ID!
            name: String!
        }

        type Post {
            id: ID!
            title: String!
        }

        type Query {
            search(q: String!): SearchResult
        }
    """

    prepare_workspace(
        tmp_path,
        """
        from sample_app.gql.api import api_gql

        SEARCH = api_gql(
            '''
            query Search($q: String!) {
                search(q: $q) {
                    ... on User {
                        __typename
                        id
                        name
                    }
                    ... on Post {
                        __typename
                        id
                        title
                    }
                }
            }
            '''
        )
        """,
        schema=schema,
    )

    with sample_app_context(tmp_path):
        changed = generate_api(tmp_path)
        assert changed is True


async def test_nullable_union_result_validation(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type Query {
            node(id: ID!): Node
        }

        union Node = User | Admin

        type User {
            id: ID!
            name: String!
        }

        type Admin {
            id: ID!
            permissions: [String!]!
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        get_node = api_gql(
            '''
            query GetNode($id: ID!) {
                node(id: $id) {
                    __typename
                    ... on User {
                        id
                        name
                    }
                    ... on Admin {
                        id
                        permissions
                    }
                }
            }
            '''
        )
    """

    def resolve_node(_root, _info, *, id: str):
        if id == "none":
            return None
        if id == "user-1":
            return {"__typename": "User", "id": id, "name": "Morty"}
        return {"__typename": "Admin", "id": id, "permissions": ["portal"]}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"node": resolve_node}},
    ) as (api, queries):
        none_result = await queries.get_node.execute(id="none")
        assert none_result.node is None

        user_result = await queries.get_node.execute(id="user-1")
        assert isinstance(user_result.node, api.GetNodeResultNodeUser)
        assert user_result.node.name == "Morty"

        admin_result = await queries.get_node.execute(id="admin-1")
        assert isinstance(admin_result.node, api.GetNodeResultNodeAdmin)
        assert admin_result.node.permissions == ["portal"]



import importlib
import json
import os
import sys
import textwrap
from collections.abc import Callable
from collections.abc import Iterator
from collections.abc import Mapping
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType

import graphql
from pydantic import alias_generators
from pytest_httpserver import HTTPServer
from werkzeug import Response

from iron_gql.generator import generate_gql_package

Resolver = Callable[..., object]
Resolvers = Mapping[str, Mapping[str, Resolver]]


def write_file(path: Path, content: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content).lstrip("\n"), encoding="utf-8")


def prepare_workspace(tmp_path: Path, query_source: str, schema: str) -> None:
    write_file(tmp_path / "schema.graphql", schema)
    write_file(tmp_path / "sample_app/__init__.py", "")
    write_file(tmp_path / "sample_app/gql/__init__.py", "")
    write_file(
        tmp_path / "sample_app/settings.py",
        "GRAPHQL_URL = 'http://testserver/graphql/'\n",
    )
    write_file(tmp_path / "sample_app/queries.py", query_source)


def clear_sample_app_modules() -> None:
    for module_name in list(sys.modules):
        if module_name == "sample_app" or module_name.startswith("sample_app."):
            sys.modules.pop(module_name, None)


@contextmanager
def import_path(path: Path) -> Iterator[None]:
    sys.path.insert(0, str(path))
    importlib.invalidate_caches()
    try:
        yield
    finally:
        sys.path.remove(str(path))
        importlib.invalidate_caches()


@contextmanager
def working_directory(path: Path) -> Iterator[None]:
    current = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(current)


@contextmanager
def sample_app_context(path: Path) -> Iterator[None]:
    with import_path(path), working_directory(path):
        yield


def generate_api(tmp_path: Path) -> bool:
    return generate_gql_package(
        schema_path=tmp_path / "schema.graphql",
        package_full_name="sample_app.gql.api",
        base_url_import="sample_app.settings:GRAPHQL_URL",
        scalars={"ID": "builtins:str"},
        to_camel_fn_full_name="pydantic.alias_generators:to_camel",
        to_snake_fn=alias_generators.to_snake,
        src_path=tmp_path,
    )


def generate_and_import(tmp_path: Path) -> tuple[ModuleType, ModuleType]:
    changed = generate_api(tmp_path)
    assert changed is True
    clear_sample_app_modules()
    api_module = importlib.import_module("sample_app.gql.api")
    queries_module = importlib.import_module("sample_app.queries")
    return api_module, queries_module


def build_schema(
    schema: str, resolvers: Resolvers | None = None
) -> graphql.GraphQLSchema:
    schema_obj = graphql.build_schema(schema)
    if not resolvers:
        return schema_obj
    for type_name, fields in resolvers.items():
        gql_type = schema_obj.get_type(type_name)
        assert isinstance(gql_type, graphql.GraphQLObjectType)
        for field_name, resolver in fields.items():
            gql_type.fields[field_name].resolve = resolver
    return schema_obj


def _setup_httpserver(httpserver: HTTPServer, schema_obj: graphql.GraphQLSchema) -> str:
    def graphql_handler(request):
        payload = request.get_json(silent=True) or {}
        result = graphql.graphql_sync(
            schema_obj,
            payload.get("query", ""),
            variable_values=payload.get("variables"),
            operation_name=payload.get("operationName"),
        )
        return Response(
            json.dumps(result.formatted),
            status=200,
            mimetype="application/json",
        )

    httpserver.expect_request("/graphql/", method="POST").respond_with_handler(
        graphql_handler
    )
    return httpserver.url_for("/graphql/")


@contextmanager
def setup_test_server(
    tmp_path: Path,
    httpserver: HTTPServer,
    schema: str,
    query_source: str,
    resolvers: Resolvers,
) -> Iterator[tuple[ModuleType, ModuleType]]:
    prepare_workspace(tmp_path, query_source, schema)
    with sample_app_context(tmp_path):
        api_module, queries_module = generate_and_import(tmp_path)
        schema_obj = build_schema(schema, resolvers)
        api_module.API_CLIENT.base_url = _setup_httpserver(httpserver, schema_obj)
        yield api_module, queries_module



import re
from pathlib import Path

from pytest_httpserver import HTTPServer

from iron_gql.conftest import setup_test_server


async def test_inline_fragment_without_type_condition(
    tmp_path: Path, httpserver: HTTPServer
):
    schema = """
        type Query {
            viewer: User!
        }

        type User {
            id: ID!
            name: String!
            email: String!
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        GET_VIEWER = api_gql(
            '''
            query GetViewer {
                viewer {
                    id
                    ... {
                        name
                        email
                    }
                }
            }
            '''
        )
    """

    def resolve_viewer(_root, _info):
        return {"id": "user-1", "name": "Morty", "email": "morty@example.com"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"viewer": resolve_viewer}},
    ) as (_, queries):
        result = await queries.GET_VIEWER.execute()
        assert result.viewer.id == "user-1"
        assert result.viewer.name == "Morty"
        assert result.viewer.email == "morty@example.com"


async def test_named_fragments(tmp_path: Path, httpserver: HTTPServer):
    schema = """
        type User {
            id: ID!
            name: String!
            email: String
        }

        type Query {
            user(id: ID!): User
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        user_fragment = api_gql(
            '''
            fragment UserFields on User {
                id
                name
            }
            '''
        )

        get_user = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                    email
                }
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {"id": id, "name": "Morty", "email": "morty@example.com"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user}},
    ) as (_, queries):
        result = await queries.get_user.execute(id="u-1")
        assert result.user is not None
        assert result.user.id == "u-1"
        assert result.user.name == "Morty"
        assert result.user.email == "morty@example.com"


async def test_duplicate_fragment_names_use_local_definition(
    tmp_path: Path, httpserver: HTTPServer
):
    schema = """
        type User {
            id: ID!
            name: String!
        }

        type Query {
            user(id: ID!): User
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        get_user = api_gql(
            '''
            fragment UserFields on User {
                id
                name
            }

            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                }
            }
            '''
        )

        other_fragment = api_gql(
            '''
            fragment UserFields on User {
                id
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {"id": id, "name": "Morty"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user}},
    ) as (_, queries):
        result = await queries.get_user.execute(id="u-1")
        assert result.user is not None
        assert result.user.id == "u-1"
        assert result.user.name == "Morty"


async def test_fragment_validation_scoped_to_query(
    tmp_path: Path, httpserver: HTTPServer
):
    schema = """
        type User {
            id: ID!
            name: String!
        }

        type Post {
            id: ID!
            title: String!
        }

        type Query {
            user(id: ID!): User
            post(id: ID!): Post
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        user_fragment = api_gql(
            '''
            fragment UserFields on User {
                id
                name
            }
            '''
        )

        post_fragment = api_gql(
            '''
            fragment PostFields on Post {
                id
                title
            }
            '''
        )

        get_user = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                }
            }
            '''
        )

        get_post = api_gql(
            '''
            query GetPost($id: ID!) {
                post(id: $id) {
                    ...PostFields
                }
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {"id": id, "name": "Morty"}

    def resolve_post(_root, _info, *, id: str):
        return {"id": id, "title": "GraphQL 101"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user, "post": resolve_post}},
    ) as (_, queries):
        user_result = await queries.get_user.execute(id="u-1")
        assert user_result.user is not None
        assert user_result.user.id == "u-1"
        assert user_result.user.name == "Morty"

        post_result = await queries.get_post.execute(id="p-1")
        assert post_result.post is not None
        assert post_result.post.id == "p-1"
        assert post_result.post.title == "GraphQL 101"


async def test_inline_fragment_definitions_not_duplicated(
    tmp_path: Path, httpserver: HTTPServer
):
    schema = """
        type User {
            id: ID!
            name: String!
        }

        type Query {
            user(id: ID!): User
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        get_user = api_gql(
            '''
            fragment UserFields on User {
                id
                name
            }

            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                }
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {"id": id, "name": "Morty"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user}},
    ) as (_, queries):
        result = await queries.get_user.execute(id="u-1")
        assert result.user is not None
        assert result.user.id == "u-1"
        assert result.user.name == "Morty"


async def test_transitive_fragments(tmp_path: Path, httpserver: HTTPServer):
    """Test that nested fragment deps (A → B → C) are resolved correctly."""
    schema = """
        type User {
            id: ID!
            name: String!
            email: String
            role: String!
        }

        type Query {
            user(id: ID!): User
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        fragment_c = api_gql(
            '''
            fragment RoleFields on User {
                role
            }
            '''
        )

        fragment_b = api_gql(
            '''
            fragment ContactFields on User {
                email
                ...RoleFields
            }
            '''
        )

        fragment_a = api_gql(
            '''
            fragment UserFields on User {
                id
                name
                ...ContactFields
            }
            '''
        )

        get_user = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                }
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {
            "id": id,
            "name": "Morty",
            "email": "morty@example.com",
            "role": "admin",
        }

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user}},
    ) as (_, queries):
        result = await queries.get_user.execute(id="u-1")
        assert result.user is not None
        assert result.user.id == "u-1"
        assert result.user.name == "Morty"
        assert result.user.email == "morty@example.com"
        assert result.user.role == "admin"


async def test_exec_source_contains_expanded_fragments(
    tmp_path: Path, httpserver: HTTPServer
):
    """Verify that the request string contains expanded fragment definitions."""
    schema = """
        type User {
            id: ID!
            name: String!
        }

        type Query {
            user(id: ID!): User
        }
    """

    query_source = """
        from sample_app.gql.api import api_gql

        user_fragment = api_gql(
            '''
            fragment UserFields on User {
                id
                name
            }
            '''
        )

        get_user = api_gql(
            '''
            query GetUser($id: ID!) {
                user(id: $id) {
                    ...UserFields
                }
            }
            '''
        )
    """

    def resolve_user(_root, _info, *, id: str):
        return {"id": id, "name": "Morty"}

    with setup_test_server(
        tmp_path,
        httpserver,
        schema,
        query_source,
        {"Query": {"user": resolve_user}},
    ) as (_, queries):
        # Read the generated code from file
        generated_code = (tmp_path / "sample_app/gql/api.py").read_text()

        # Verify the generated code contains the fragment definition in the request
        assert "fragment UserFields on User" in generated_code
        # The fragment should appear as part of the gql.gql() call
        assert re.search(r"gql\.gql\(['\"].*fragment UserFields", generated_code)

        # Also verify execution works
        result = await queries.get_user.execute(id="u-1")
        assert result.user is not None
        assert result.user.id == "u-1"
